__version__ = "0.0.2"
__title__ = "delimited"
__summary__ = "Defines types that allow for accessing and modifying nested data"
__url__ = "https://github.com/chrisantonellis/delimited"
__author__ = "Christopher Antonellis"
__email__ = "christopher.antonellis@gmail.com"
__license__ = "MIT License"

from delimited.container import NestedDict
from delimited.container import NestedOrderedDict
from delimited.container import DelimitedDict
from delimited.container import DelimitedOrderedDict
